# -*- coding: iso-8859-1 -*-
# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

# $Id: as_setup.py,v 1.25 2005/06/30 06:38:47 mcourtoi Exp $
# $Name:  $

"""This module defines following classes :
- SETUP        main class to configure, make, install Code_Aster products,
- DEPENDENCIES stores dependencies between products,
- SUMMARY      stores setup informations,
- SYSTEM       encapsulates calls to system commands,
- FIND_TOOLS   package of utilities to find files, libraries...

+ exceptions to control errors occurences.
"""

__all__ = ['SETUP', 'SUMMARY', 'SYSTEM', 'DEPENDENCIES', 'FIND_TOOLS',
   'SetupCheckError', 'SetupChgFilesError', 'SetupChmodError', 'SetupCleanError',
   'SetupConfigureError', 'SetupError', 'SetupExtractError', 'SetupInstallError',
   'SetupMakeError',]

temporay_folder='/tmp'

import sys
import os
import os.path
import popen2
import glob
import re
import time
import traceback
import tarfile
import compileall
from types import *
EnumTypes=(ListType, TupleType)

# ----- differ messages translation
def _(mesg): return mesg

class SetupError(Exception):           pass
class SetupExtractError(SetupError):   pass
class SetupConfigureError(SetupError): pass
class SetupChgFilesError(SetupError):  pass
class SetupChmodError(SetupError):     pass
class SetupMakeError(SetupError):      pass
class SetupInstallError(SetupError):   pass
class SetupCheckError(SetupError):     pass
class SetupCleanError(SetupError):     pass

def _chgline(line, dtrans, delimiter='\?'):
   """Change all strings by their new value provided by 'dtrans' dictionnary
   in the string 'line'.
   """
   for old, new in dtrans.items():
      line=re.sub(delimiter+old+delimiter, str(new), line)
   return line

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class SETUP:
   """
   Attributes :
    .product      : product name,
    .version      : product version,
    .description  : description of the product,
    .archive      : archive filename (gzip and bzip2 compressed archives support
      is provided by tarfile module),
    .content      : name of the directory contained in archive ('.' else),
    .verbose      : True for verbose mode,
    .debug        : True for debug mode,
    .installdir   : installation directory,
    .sourcedir    : directory where is 'archive',
    .workdir      : directory used for extraction, compilation...
    .delimiter    : ChgFiles search strings enclosed by this delimiter
      (note with special character in regexp, default is '\?'),
    .manual       : if False, Go() is automatically called by __init__
      (NOTE : if True and error occurs SETUP instance won't be created because
       __init__ wasn't finish),
    .actions      : give the list of actions run by 'Go' method and arguments,
    .exit_code    : None during process, exit code after,
    
    .depend       : DEPENDENCIES object.

     + Shell, VerbStart, VerbIgnore, VerbEnd : shortcuts to SYSTEM methods

   NB : Configure, Make, Install and Check methods can be provided by 'user'
      through 'external' argument for particular applications.
   """
   _separ='\n' + '-'*80 + '\n'
   _fmt_info=_separ+_(""" Installation of   : %s %s
   %s
 Archive filename  : %s
 Destination       : %s
 Working directory : %s""")+_separ
   _fmt_enter=_("entering directory '%s'")
   _fmt_leave=_("leaving directory '%s'")
   _fmt_title='\n >>> %s <<<\n'
   _fmt_ended=_separ+_(' Installation of %s %s successfully completed')+_separ
#-------------------------------------------------------------------------------
   def __init__(self, **kargs):
      """Initializes an instance of an SETUP object and calls Info to print
      informations.
      """
      self.exit_code=None
      # product parameters
      required_args=('product', 'version', 'description', 'archive', 'content',
                     'installdir', 'sourcedir')
      for arg in required_args:
         try:
            setattr(self, arg, kargs[arg])
         except KeyError, msg:
            raise SetupError, _('missing argument %s') % str(msg)
      optionnal_args={
         'depend'    : None,
         'workdir'   : os.path.join(temporay_folder,
               'install_'+self.product+'.'+str(os.getpid())),
         'delimiter' : '\?',
         'manual'    : True,
      }
      for arg, default in optionnal_args.items():
         setattr(self, arg, kargs.get(arg, default))

      # default actions can only be set after initialisation
      default_actions = (
         ('Extract'   , {}),
         ('Configure' , {}),
         ('Make'      , {}),
         ('Install'   , {}),
         ('Check'     , {}),
         ('Clean'     , {}),
      )
      self.actions = kargs.get('actions', default_actions)
      for d in ('installdir', 'sourcedir', 'workdir'):
         setattr(self, d, os.path.abspath(getattr(self, d)))

      # external methods
      system=kargs.get('system')
      if system==None:
         raise SetupError, _("Argument not found 'system'")
      self.verbose    = system.verbose
      self.debug      = system.debug
      self.Shell      = system.local_shell
      self.VerbStart  = system.VerbStart
      self.VerbIgnore = system.VerbIgnore
      self.VerbEnd    = system.VerbEnd
      self._print = kargs['log']._print
      
      # print product informations
      self.Info()

      # Go if not manual
      if not self.manual:
         self.Go()

#-------------------------------------------------------------------------------
   def _call_external(self, **kargs):
      """Call an external user's method to perform an operation.
      """
      try:
         kargs.get('external')(self, **kargs)
      except:
         self.exit_code=4
         self._print(self._separ, term='')
         self._print(self._fmt_title % _('External method traceback'))
         traceback.print_exc()
         self._print(self._separ)
         raise SetupConfigureError, _("external method failed (see traceback above).")

#-------------------------------------------------------------------------------
   def Go(self):
      """Run successively each actions defined in 'self.actions',
      call PreCheck() first.
      """
      self.PreCheck()
      self.depend.LastCheck()
      for act, kargs in self.actions:
         if not hasattr(self, act):
            raise SetupError, _("unknown action '%s'") % act
         else:
            getattr(self, act)(**kargs)
      self.exit_code=0
      self._print(self._fmt_ended % (self.product, self.version))

#-------------------------------------------------------------------------------
   def PreCheck(self, check_values=True):
      """Check for permission and variables settings.
      """
      # fill cache file
      self.depend.FillCache(self.product)
      
      # check for permissions and create directories
      self.VerbStart(_('Checking permissions...'))
      iret=0
      ldirs=[self.installdir, self.workdir]
      for d in ldirs:
         try:
            if not os.path.exists(d):
               os.makedirs(d)
            elif os.access(d, os.W_OK)==0:
               raise OSError, _('no permission to write in %s') % d
         except OSError:
            iret+=1
            if iret==1: self._print()
            self._print(_(' Unsufficient permission to create or write in %s') % repr(d))
      if iret<>0: self.VerbStart('')   # just for pretty print if fails
      self.VerbEnd(iret)
      if iret<>0:
         raise SetupError, _('permission denied')

#-------------------------------------------------------------------------------
   def Info(self):
      """Print informations about the installation of current product
      """
      self._print(self._fmt_info % (self.product, self.version, self.description,
                              self.archive, self.installdir, self.workdir))

#-------------------------------------------------------------------------------
   def Extract(self, **kargs):
      """Extract the archive of the product into 'workdir'.
         archive : full pathname of archive,
         command : alternative command line to extract archive (MUST include
            archive filename !),
         extract_as : rename content.
      """
      self._print(self._fmt_title % _('Extraction'))
      if kargs.get('external')<>None:
         self._call_external(**kargs)
         return
      archive = kargs.get('archive', os.path.join(self.sourcedir,self.archive))
      command = kargs.get('command')
      newname = kargs.get('extract_as', None)
      path=self.workdir
      iextr_as=newname<>None and self.content<>newname and self.content<>'.'
      if iextr_as:
         True
         path=os.path.join(self.workdir,'tmp_extract')
         if not os.path.exists(path):
            os.makedirs(path)

      if not os.path.isfile(archive):
         self.exit_code=1
         raise SetupExtractError, "file not found : "+archive

      prev=os.getcwd()
      self._print(self._fmt_enter % path)
      os.chdir(path)

      if command<>None:
         iret,output=self.Shell(command,
               alt_comment=_('Extracting %s...') % os.path.basename(archive))
         if iret<>0:
            self.exit_code=4
            raise SetupExtractError, _('error during extracting archive %s') % archive
      else:
         # extract archive using tarfile module
         if not tarfile.is_tarfile(archive):
            self.exit_code=2
            raise SetupExtractError, _('unsupported archive format for %s') % archive
         else:
            self.VerbStart(_('Extracting %s...') % os.path.basename(archive))
            iret=0
            n=0
            try:
               tar=tarfile.open(archive,'r')
               tar.errorlevel=2
               for ti in tar:
                  n+=1
                  tar.extract(ti)
            except tarfile.ExtractError:
               iret=4
            except IOError:
               iret=8
            self.VerbEnd(iret)
            tar.close()
            self._print(_(' --- %d files extracted') % n)
            if iret<>0:
               self.exit_code=iret
               raise SetupExtractError, _('error during extracting archive %s') % archive
      
      if iextr_as:
         iret=0
         self.VerbStart(_('Renaming %s to %s...') % (self.content, newname))
         newname=os.path.join(self.workdir,newname)
         try:
            if os.path.exists(newname):
               self._print()
               iret,output=self.Shell('rm -rf '+newname,
                     alt_comment=_('Deleting previous content of %s...') % newname)
               if iret<>0:
                  raise OSError
            os.rename(self.content, newname)
         except (OSError, IOError), msg:
            iret=4
            self._print()
            raise SetupExtractError, _('error renaming %s to %s') \
                  % (self.content, newname)
         self.VerbEnd(iret)

      self._print(self._fmt_leave % path)
      os.chdir(prev)
      if iextr_as:
         self.Clean(to_delete=path)

#-------------------------------------------------------------------------------
   def Configure(self, **kargs):
      """Configuration of the product.
         command : alternative command line
         path    : directory to build
      """
      self._print(self._fmt_title % _('Configuration'))
      if kargs.get('external')<>None:
         self._call_external(**kargs)
         return
      command = kargs.get('command')
      path    = kargs.get('path', os.path.join(self.workdir,self.content))
      if command==None:
         command='./configure --prefix='+self.installdir
      
      if not os.path.isdir(path):
         self.exit_code=4
         raise SetupConfigureError, _('directory not exists %s') % path

      prev=os.getcwd()
      self._print(self._fmt_enter % path)
      os.chdir(path)

      iret,output=self.Shell(command, follow_output=self.verbose,
            alt_comment=_('configure %s installation...') % self.product)
      if iret<>0:
         if not self.verbose: self._print(output)
         self.exit_code=4
         raise SetupConfigureError, _('error during configure')

      self._print(self._fmt_leave % path)
      os.chdir(prev)
      
#-------------------------------------------------------------------------------
   def ChgFiles(self, **kargs):
      """Modify files from 'files' according a dictionnary 'dtrans'
         defined as { 'string_to_replace' : 'new_value' }.
      Pathnames are relative to workdir/content or to 'path' argument.
      """
      self._print(self._fmt_title % _('Set parameters to their value'))
      if kargs.get('external')<>None:
         self._call_external(**kargs)
         return
      chglist = kargs.get('files', [])
      dtrans  = kargs.get('dtrans',{})
      delimiter=kargs.get('delimiter', self.delimiter)
      path=kargs.get('path', os.path.join(self.workdir,self.content))

      if not os.path.isdir(path):
         self.exit_code=4
         raise SetupChgFilesError, _('directory not exists %s') % path

      prev=os.getcwd()
      self._print(self._fmt_enter % path)
      os.chdir(path)

      for f0 in chglist:
         for f in glob.glob(f0):
            iret=0
            self.VerbStart(_('modifying %s') % f)
            if os.path.isfile(f):
               iret=self._chgone(f, dtrans, delimiter)
               self.VerbEnd(iret)
            else:
               self.VerbIgnore()

      self._print(self._fmt_leave % path)
      os.chdir(prev)

   def _chgone(self, filename, dtrans, delimiter=None):
      """Change all strings by their new value provided by 'dtrans' dictionnary
      in the existing file 'filename'.
      """
      if delimiter==None:
         delimiter=self.delimiter
      iret=0
      try:
         os.rename(filename, filename+'.orig')
      except OSError:
         return 4
      fsrc=open(filename+'.orig', 'r')
      fnew=open(filename, 'w')
      for line in fsrc:
         fnew.write(_chgline(line, dtrans, delimiter))
      fnew.close()
      fsrc.close()
      os.remove(filename+'.orig')
      return iret

#-------------------------------------------------------------------------------
   def Make(self, **kargs):
      """Compilation of product.
         command : alternative command line
         path    : directory to build (or list of directories)
      """
      self._print(self._fmt_title % _('Build the product'))
      if kargs.get('external')<>None:
         self._call_external(**kargs)
         return
      command = kargs.get('command')
      lpath    = kargs.get('path', os.path.join(self.workdir,self.content))
      if not type(lpath) in EnumTypes:
         lpath=[lpath,]
      if command==None:
         command='make'
      
      for path in lpath:
         if not os.path.isdir(path):
            self.exit_code=4
            raise SetupConfigureError, _('directory not exists %s') % path

         prev=os.getcwd()
         self._print(self._fmt_enter % path)
         os.chdir(path)

         iret,output=self.Shell(command, follow_output=self.verbose,
               alt_comment=_('compiling %s...') % self.product)
         if iret<>0:
            if not self.verbose: self._print(output)
            self.exit_code=4
            raise SetupMakeError, _('error during compilation')

         self._print(self._fmt_leave % path)
         os.chdir(prev)

#-------------------------------------------------------------------------------
   def Install(self, **kargs):
      """Perform installation of the product.
         command : alternative command line
         path    : directory to build
      """
      self._print(self._fmt_title % _('Installation'))
      if kargs.get('external')<>None:
         self._call_external(**kargs)
         return
      command = kargs.get('command')
      path    = kargs.get('path', os.path.join(self.workdir,self.content))
      if command==None:
         command='make install'
      
      if not os.path.isdir(path):
         self.exit_code=4
         raise SetupInstallError, _('directory not exists %s') % path

      prev=os.getcwd()
      self._print(self._fmt_enter % path)
      os.chdir(path)

      iret,output=self.Shell(command, follow_output=self.verbose,
            alt_comment=_('installing %s to %s...') % (self.product, self.installdir))
      if iret<>0:
         if not self.verbose: self._print(output)
         self.exit_code=4
         raise SetupInstallError, _('error during installation')

      self._print(self._fmt_leave % path)
      os.chdir(prev)

#-------------------------------------------------------------------------------
   def PyCompile(self, **kargs):
      """Recursively descend the directory tree named by 'path', compiling
      all .py files along the way.
         path    : one or more directories
      """
      self._print(self._fmt_title % _('Compiling Python source files'))
      lpath    = kargs.get('path', os.path.join(self.workdir,self.content))
      if not type(lpath) in EnumTypes:
         lpath=[lpath,]

      iret=0
      for path in lpath:      
         if not os.path.isdir(path):
            self.exit_code=4
            raise SetupInstallError, _('directory not exists %s') % path

         prev=os.getcwd()
         self._print(self._fmt_enter % path)
         os.chdir(path)

         ierr=0
         self.VerbStart(_('Building byte-code files from %s...') % path)
         try:
            compileall.compile_dir(path, quiet=(self.verbose==False))
         except Exception, msg:
            ierr=iret=4
            self._print(msg)
         if self.verbose: self.VerbStart('')   # just for pretty print
         self.VerbEnd(ierr)
      if iret<>0:
         raise SetupInstallError, _('error during byte-code built')

      self._print(self._fmt_leave % path)
      os.chdir(prev)

#-------------------------------------------------------------------------------
   def Chmod(self, **kargs):
      """Change mode of the 'files' to 'mode'.
      Pathnames are relative to installdir or to 'path' argument.
      """
      self._print(self._fmt_title % _('Set files permission'))
      if kargs.get('external')<>None:
         self._call_external(**kargs)
         return
      chglist = kargs.get('files', [])
      try:
         mode    = int(kargs.get('mode', 0644))
      except ValueError:
         self.exit_code=4
         raise SetupChmodError, _("an integer is required for 'mode'")
      path=kargs.get('path', self.installdir)

      if not os.path.isdir(path):
         self.exit_code=4
         raise SetupChgFilesError, _('directory not exists %s') % path

      prev=os.getcwd()
      self._print(self._fmt_enter % path)
      os.chdir(path)

      for f0 in chglist:
         for f in glob.glob(f0):
            iret=0
            self.VerbStart(_('applying chmod %04o to %s') % (mode,f))
            if os.path.isfile(f):
               try:
                  os.chmod(f, mode)
               except OSError:
                  iret=4
               self.VerbEnd(iret)
            else:
               self.VerbIgnore()

      self._print(self._fmt_leave % path)
      os.chdir(prev)

#-------------------------------------------------------------------------------
   def Check(self, **kargs):
      """Check if installation was successfull.
         path    : directory to build
      """
      self._print(self._fmt_title % _('Check for installation'))
      if kargs.get('external')<>None:
         self._call_external(**kargs)
         return
      command = kargs.get('command')
      path    = kargs.get('path', os.path.join(self.workdir,self.content))
      if command==None:
         command='make check'
      
      if not os.path.isdir(path):
         self.exit_code=4
         raise SetupConfigureError, _('directory not exists %s') % path

      prev=os.getcwd()
      self._print(self._fmt_enter % path)
      os.chdir(path)

      iret,output=self.Shell(command, follow_output=self.verbose,
            alt_comment=_('checking %s installation...') % self.product)
      if iret<>0:
         if not self.verbose: self._print(output)
         self.exit_code=4
         raise SetupCheckError, _('error during checking installation')

      self._print(self._fmt_leave % path)
      os.chdir(prev)

#-------------------------------------------------------------------------------
   def Clean(self, **kargs):
      """Clean working directory.
         'to_delete' : list of objects to delete
      Pathnames are relative to workdir or to 'path' argument.
      """
      self._print(self._fmt_title % _('Clean temporary objects'))
      to_del = kargs.get('to_delete', [self.content])
      path   = kargs.get('path', self.workdir)
      if not type(to_del) in EnumTypes:
         to_del=[to_del]

      prev=os.getcwd()
      self._print(self._fmt_enter % path)
      os.chdir(path)

      for obj in [os.path.abspath(o) for o in to_del]:
         iret,output=self.Shell(cmd='rm -rf '+obj,
               alt_comment=_('deleting %s...') % obj)

      self._print(self._fmt_leave % path)
      os.chdir(prev)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class DEPENDENCIES:
   """Class to store dependencies between products.
   Attributes :
      prod_req : dict where keys are products names and values the list of
         variables required by them,
      req_prod : reverse of prod_req,
      prod_set : dict where keys are products names and values the list of
         variables set by them,
      set_prod : reverse of prod_set,
      req_obj  : dict product : required objects (only tested by LastCheck),
      cfg      : dictionnary containing parameters values,
      cache    : cache filename,
      debug    : debug mode.
   """
   _separ='\n' + '-'*80 + '\n'
#-------------------------------------------------------------------------------
   def __init__(self, **kargs):
      """Constructor
      """
      self.cfg=kargs.get('cfg', {})
      self.cache=kargs.get('cache', {})
      self._print = kargs['log']._print

      # external methods
      system=kargs.get('system')
      if system==None:
         raise SetupError, _("Argument not found 'system'")
      self.debug      = system.debug
      self.Shell      = system.local_shell
      self.VerbStart  = system.VerbStart
      self.VerbIgnore = system.VerbIgnore
      self.VerbEnd    = system.VerbEnd
      
      self.prod_req = {}
      self.req_prod = {}
      self.prod_set = {}
      self.set_prod = {}
      self.req_obj  = {}
      if kargs.get('req')<>None or kargs.get('set')<>None:
         self.Add('__main__', kargs.get('req', []), kargs.get('set', []))

#-------------------------------------------------------------------------------
   def PrintContent(self, **kargs):
      """Print content
      """
      self._print(self._separ, ('Content of cfg :'))
      self._print(self.cfg)
      self._print(_('\nProducts prerequisites :'))
      self._print(self.prod_req)
      self._print(_('\nProducts which require these variables :'))
      self._print(self.req_prod)
      self._print(_('\nVariables set by products :'))
      self._print(self.prod_set)
      self._print(_('\nProducts which set these variables :'))
      self._print(self.set_prod)

#-------------------------------------------------------------------------------
   def Add(self, product, req=[], set=[], reqobj=[]):
      """Add dependencies of 'product'.
      'req' should contains all variables set before this product setup.
      Required variables that could be deduced from others should only be
      in 'set'.
      'reqobj' contains names of objects (actually only files) required.
      """
      def check_varname(v):
         d={}
         try:
            exec(v+'=0') in d
         except Exception, msg:
            raise SetupError, _(' invalid variable name %s' % repr(v))
      
      if not self.prod_req.has_key(product): self.prod_req[product]=[]
      if not self.prod_set.has_key(product): self.prod_set[product]=[]
      if not self.req_obj.has_key(product):  self.req_obj[product]=[]
      
      # takes from req only variables not set by product
      for r in [r for r in req if not r in set]:
         check_varname(r)
         self.prod_req[product].append(r)
         if not self.req_prod.has_key(r): self.req_prod[r]=[]
         self.req_prod[r].append(product)
      for s in set:
         check_varname(s)
         self.prod_set[product].append(s)
         if not self.set_prod.has_key(s): self.set_prod[s]=[]
         self.set_prod[s].append(product)

      for o in reqobj:
         self.req_obj[product].append(o)

      # check
      self.CheckDepVal(product)

      # set variables are considered as required
      for r in set:
         check_varname(r)
         self.prod_req[product].append(r)
         if not self.req_prod.has_key(r): self.req_prod[r]=[]
         self.req_prod[r].append(product)

#-------------------------------------------------------------------------------
   def CheckDepVal(self, product='all'):
      """Check for dependencies, values setting and permission.
      """
      if self.debug:
         self._print('PASSAGE CHECKDEPVAL product='+product, self._separ, term='')
         self.PrintContent()

      self._print(self._separ, term='')
      self.VerbStart(_('Checking for dependencies and required ' \
            'variables for %s...') % repr(product))
      iret, lerr = self.Check(product)
      if iret<>0: self.VerbStart('')   # just for pretty print if fails
      self.VerbEnd(iret)
      if iret<>0:
        raise SetupError, _('inconsistent dependencies or missing variables'\
               +'\n     Problem with : %s') % ', '.join(lerr)

#-------------------------------------------------------------------------------
   def Check(self, product='all', check_deps=True, check_values=True):
      """Check for dependencies.
      """
      iret=0
      lerr=[]
      if product=='all':
         product=self.prod_req.keys()
      if not type(product) in EnumTypes:
         product=[product,]
      for p in product:
         if not self.prod_req.has_key(p):
            iret=-1
            self._print()
            self._print(_(' No dependencies found for %s') % repr(p))
         else:
            for v in self.prod_req[p]:
               err=0
               if check_values and not self.cfg.has_key(v):
                  iret+=1
                  err=1
                  lerr.append(v)
                  if iret==1: self._print()
                  self._print(_(' %15s is required by %s but not set.') % (v, repr(p)))
               if check_deps and not v in self.set_prod.keys() and err==0:
                  iret+=1
                  lerr.append(v)
                  if iret==1: self._print()
                  self._print(_(' %15s is required by %s') % (v, repr(p)))
      return iret, lerr

#-------------------------------------------------------------------------------
   def LastCheck(self, product='all'):
      """Check for required objects are present.
      """
      iret=0
      lerr=[]
      if product=='all':
         product=self.req_obj.keys()
      if not type(product) in EnumTypes:
         product=[product,]
      for p in product:
         if not self.req_obj.has_key(p):
            iret=-1
            self._print()
            self._print(_(' No objects dependencies found for %s') % repr(p))
         else:
            for v in self.req_obj[p]:
               typ=v.split(':')[0]
               val=''.join(v.split(':')[1:])
               if typ=='file':
                  vf=_chgline(val, self.cfg)
                  if not os.path.exists(vf):
                     iret+=1
                     lerr.append(vf)
                     self._print(_(' %s requires this file : %s') % (repr(p), vf))
               else:
                  raise SetupError, _('unknown type of object : %s') % typ
      if iret>0:
        raise SetupError, _('missing objects'\
               +'\n     Problem with : %s') % ', '.join(lerr)

#-------------------------------------------------------------------------------
   def FillCache(self, product='all'):
      """Fill cache file with values set by product.
      NOTE : call it only if you are sure variables are set !
      """
      lerr=[]
      if product=='all':
         product=self.prod_req.keys()
      if not type(product) in EnumTypes:
         product=[product,]
      # fill cache file with values set by product(s)
      self.VerbStart(_('Filling cache...'))
      iret=0
      f=open(self.cache, 'a')
      for p in product:
         f.write('\n# Variables set by %s at %s\n' % \
               (repr(p),time.strftime('%c')))
         # alphabetical sort (easier to read)
         self.prod_set[p].sort()
         for v in self.prod_set[p]:
            if self.cfg.has_key(v):
               f.write('%s = %s\n' % (v, repr(self.cfg[v])))
            else:
               iret=255
               self._print()
               lerr.append(v)
               self._print(_(' %15s not yet defined' % repr(v)))
      f.close()
      if iret<>0: self.VerbStart('')   # just for pretty print if fails
      self.VerbEnd(iret)
      if iret<>0:
         raise SetupError, _('unavailable variables' \
               +'\n     Problem with : %s') % ', '.join(lerr)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class SUMMARY:
   """This class collects informations about product installations as
   diagnostic, exception raised if any...
   """
   _separ='\n' + '-'*80 + '\n'
   _fmt_title=_separ + '      %s' + _separ
   _fmt_sum=_(""" Installation of   : %(product)s %(version)s
 Destination       : %(installdir)s
 Elapsed time      : %(time).2f s""")
   _fmt_except=_('\n *** Exception %s raised : %s\nSee detailed traceback in' \
         ' the logfile')
#-------------------------------------------------------------------------------
   def __init__(self, list_of_products, **kargs):
      self.products=list_of_products
      self.diag={}

      # external methods
      system=kargs.get('system')
      self._print = kargs['log']._print
      if system==None:
         raise SetupError, _("Argument not found 'system'")
      self.Shell      = system.local_shell
      self.VerbStart  = system.VerbStart
      self.VerbIgnore = system.VerbIgnore
      self.VerbEnd    = system.VerbEnd
      
      for p in self.products:
         self.diag[p]={
            'product'      : p,
            'version'      : '(version unavailable)',
            'installdir'   : 'unknown',
            'exit_code'    : None,
            'time'         : 0.,
            'tb_info'      : None,
         }

#-------------------------------------------------------------------------------
   def Set(self, product, setup, dt, traceback_info=None):
      """Set informations about a product installation.
      """
      if isinstance(setup, SETUP):
         self.diag[product].update({
            'version'      : setup.version,
            'installdir'   : setup.installdir,
            'exit_code'    : setup.exit_code,
         })
      else:
         self.diag[product]['exit_code']=4
      self.diag[product]['time']=dt
      if traceback_info<>None:
         self.diag[product]['tb_info']=traceback_info
      sys.exc_clear()

#-------------------------------------------------------------------------------
   def Print(self):
      """Return a representation of the SUMMARY object
      """
      self._print(self._fmt_title % _('SUMMARY OF INSTALLATION'))
      for p in self.products:
         self.VerbStart(self._fmt_sum % self.diag[p])
         if self.diag[p]['exit_code']==None:
            self.VerbIgnore()
         else:
            if self.diag[p]['exit_code']<>0:
               self._print(self._fmt_except % self.diag[p]['tb_info'][:2])
#                traceback.print_tb(self.diag[p]['tb_info'][2])
               self.VerbStart('')   # just for pretty print if fails
            self.VerbEnd(self.diag[p]['exit_code'])

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class SYSTEM:
   """Class to encapsultate "system" commands (this a simplified version of
   ASTER_SYSTEM class defined in ASTK_SERV part).
   """
   # this value should be set during installation step.
   MaxCmdLen=1024
   # line length -9
   _LineLen=80-9
#-------------------------------------------------------------------------------
   def __init__(self, run, **kargs):
      """run : dictionnary to define 'verbose', 'debug'
      """
      self.verbose   = run['verbose']
      self.debug     = run['debug']
      self._print = kargs['log']._print
      if kargs.has_key('maxcmdlen'):
         self.MaxCmdLen = kargs['maxcmdlen']

#-------------------------------------------------------------------------------
   def _mess(self,msg,cod=''):
      """Just print a message
      """
      self._print('%-18s %s' % (cod,msg))

#-------------------------------------------------------------------------------
   def VerbStart(self,cmd,verbose=None):
      """Start message in verbose mode
      """
      Lm=self._LineLen
      if verbose==None:
         verbose=self.verbose
      if verbose:
         pcmd=cmd
         if len(cmd)>Lm-2 or cmd.count('\n')>0:
            pcmd=pcmd+'\n'+' '*Lm
         self._print(('%-'+str(Lm)+'s') % (pcmd,), term='')
         sys.stdout.flush()

#-------------------------------------------------------------------------------
   def VerbEnd(self,iret,output='',verbose=None):
      """Ends message in verbose mode
      """
      if verbose==None:
         verbose=self.verbose
      if verbose:
         if iret==0:
            self._print('[  OK  ]')
         else:
            self._print(_('[FAILED]'))
            self._print(_('Exit code : %d') % iret)
         if (iret<>0 or self.debug) and output:
            self._print(output)

#-------------------------------------------------------------------------------
   def VerbIgnore(self,verbose=None):
      """Ends message in verbose mode
      """
      if verbose==None:
         verbose=self.verbose
      if verbose:
         self._print('[ SKIP ]')

#-------------------------------------------------------------------------------
   def local_shell(self,cmd,bg=False,verbose=None,follow_output=False,
         alt_comment=None,interact=False):
      """Execute a command shell
         cmd      : command
         bg       : put command in background if True
         verbose  : print status messages during execution if True
         follow_output : follow interactively output of command
         alt_comment : print this "alternative comment" instead of "cmd"
         interact : allow the user to interact with the process
            (don't close stdin). bg=True implies interact=False.
      Return :
         iret     : exit code if bg = False,
                    process id if bg = True
         output   : output lines (as string)
      """
      if not alt_comment:
         alt_comment=cmd
      if verbose==None:
         verbose=self.verbose
      if bg:
         interact=False
      if len(cmd)>self.MaxCmdLen:
         self._mess((_('length of command shell greater '\
               'than %d characters.') % self.MaxCmdLen),_('<A>_ALARM'))
      if self.debug:
         self._print('<DBG> <local_shell>',cmd)
         self._print('<DBG> <local_shell> background mode : ',bg)
      # exec
      self.VerbStart(alt_comment,verbose=verbose)
      if follow_output and verbose:
         self._print(_('\nCommand output :'))
      # run interactive command
      if interact:
         iret=os.system(cmd)
         return iret,''
      # use popen to manipulate stdout/stderr
      output=[]
      p=popen2.Popen4(cmd)
      p.tochild.close()
      if not bg:
         if not follow_output:
            output=p.fromchild.readlines()
         else:
            while p.poll()==-1:
               output.append(p.fromchild.readline())
               # \n already here...
               self._print(output[-1], term='')
               sys.stdout.flush()
            # to be sure to empty the buffer
            end=p.fromchild.readlines()
            self._print(''.join(end))
            output.extend(end)
         iret=p.wait()
      else:
         iret=0
      p.fromchild.close()
      output=''.join(output)

      # repeat header message
      if follow_output:
         self.VerbStart(alt_comment,verbose=verbose)
      mat=re.search('EXIT_CODE=([0-9]+)',output)
      if mat:
         try:
            iret=int(mat.group(1))
         except:
            pass
      self.VerbEnd(iret,output,verbose=verbose)
      if bg:
         iret=p.pid
         if verbose:
            self._print('Process ID : ',iret)
      return iret,output

#-------------------------------------------------------------------------------
   def GetHostName(self, host=None):
      """Return hostname of the machine 'host' or current machine if None.
      """
      from socket import gethostname, gethostbyaddr
      if host==None:
         host = gethostname()
      fqn, alias, ip = gethostbyaddr(host)
      if fqn.find('localhost')>-1:
         alias=[a for a in alias if a.find('localhost')<0]
         if len(alias)>0:
            fqn=alias[0]
         for a in alias:
            if a.find('.')>-1:
               fqn=a
               break
      return fqn

#-------------------------------------------------------------------------------
def _getsubdirs(dirs, maxdepth=5):
   """Returns the list of subdirectories of 'dirs' up to 'maxdepth'.
   Note that 'dirs' will appear at the beginning of the returned list.
   """
   if not type(dirs) in EnumTypes:
      dirs=[dirs]
   dirs=[os.path.realpath(i) for i in dirs if i<>'']
   new=[]
   for d in dirs:
      if not d in new:
         new.append(d)
   if maxdepth<1: return new
   for d in dirs:
      level=len(d.split(os.path.sep))
      for root, dirs, nondirs in os.walk(d):
         lev=len(root.split(os.path.sep))
         if lev<=level+maxdepth and not root in new:
            new.append(root)
   return new

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class FIND_TOOLS:
   """Some utilities to find files, libraries...
   """
   _fmt_search=_('Checking for %s... ')
   _ext_static='.a'
   _ext_shared='.so'
#-------------------------------------------------------------------------------
   def __init__(self, **kargs):
      self._print     = kargs.get('log', None)._print
      self.maxdepth   = kargs.get('maxdepth', 0)
      self.prefshared = kargs.get('prefshared', False)
      self.paths      = {'bin' : [], 'lib' : [], 'inc' : []}

#-------------------------------------------------------------------------------
   def find_file(self, filenames, std_dirs, paths, maxdepth=None):
      """Returns absolute path of the first of 'filenames' located
      in paths+std_dirs (so search from paths first) or None if no one was found.
      """
      if not type(filenames) in EnumTypes:
         filenames=[filenames,]
      if not type(paths) in EnumTypes:
         paths=[paths,]
      if maxdepth==None:
         maxdepth=self.maxdepth
      for name in filenames:
         self._print(self._fmt_search % name, term='')
         # Check the user's directories and then standard locations
         for dir in _getsubdirs(paths+std_dirs, maxdepth):
            f = os.path.join(dir, name)
            if os.path.exists(f):
               self._print(os.path.abspath(f))
               return os.path.abspath(f)
         self._print(_('no'))
      # Not found anywhere
      return None

#-------------------------------------------------------------------------------
   def find_and_set(self, cfg, var, filenames, std_dirs, paths=[],
                    err=True, typ='bin', append=False, maxdepth=None):
      """Uses find_files to search 'filenames' in paths+std_dirs.
      - If 'err' is True, raises SetupConfigureError if no one was found.
      - Value set in cfg[var] depends on 'typ' :
         typ='bin' : cfg[var]='absolute filename found'
         typ='inc' : cfg[var]='-Idirname'
         typ='lib' : cfg[var]='static lib' name or '-Ldirname -llib'
      - If cfg[var] already exists :
         if 'append' is False, cfg[var] is unchanged ('previously set' message)
         if 'append' is True, new found value is appended to cfg[var]
      """
      if not type(filenames) in EnumTypes:
         filenames=[filenames,]
      if maxdepth==None:
         maxdepth=self.maxdepth
      if not cfg.has_key(var) or append:
         found=self.find_file(filenames, std_dirs, paths, maxdepth)
         if found==None:
            if cfg.has_key(var) and not append:
               del cfg[var]
            elif not cfg.has_key(var) and append:
               cfg[var]=''
            if err:
               raise SetupConfigureError, _('Unable to find file %s') \
                     % ' or '.join(filenames)
         else:
            root, ext = os.path.splitext(found)
            dirname   = os.path.abspath(os.path.dirname(root))
            basename  = os.path.basename(root)
            if typ=='lib':
               if ext==self._ext_shared:
                  found='-L%s -l%s' % (dirname, re.sub('^lib', '', basename))
            elif typ=='inc':
               found='-I%s' % dirname
            if not cfg.has_key(var) or not append:
               cfg[var]=found
            else:
               cfg[var]=(cfg[var]+' '+found).strip()
            if not dirname in self.paths[typ]:
               self.paths[typ].append(dirname)
      else:
         self._print(self._fmt_search % ' or '.join(filenames), term='')
         self._print(_("%s (previously set)") % cfg[var])

#-------------------------------------------------------------------------------
   def findlib_and_set(self, cfg, var, libname, std_dirs, paths=[],
                    err=True, typ='bin', append=False, prefshared=None,
                    maxdepth=None):
      """Same as find_and_set but expands 'libname' as static and shared
      library filenames.
      """
      l_exts=[self._ext_static, self._ext_shared]
      if maxdepth==None:
         maxdepth=self.maxdepth
      if prefshared==None:
         prefshared=self.prefshared
      if prefshared:
         l_exts.reverse()
      self.find_and_set(cfg, var,
         ['lib'+libname+ext for ext in l_exts],
         std_dirs, paths, err, typ, append, maxdepth)

#-------------------------------------------------------------------------------
   def addtopath(self, typ, var):
      """Get the directory of object of type 'typ' from 'var', and add it
      to corresponding paths items.
      'var' is supposed to be a result of find_and_set.
      """
      toadd=[]
      if   typ=='bin':
         for v in var.split():
            toadd.append(os.path.dirname(v))
      elif typ=='lib':
         for v in var.split():
            if   v[:2]=='-L':
               toadd.append(v[2:])
            elif v[:2]=='-l':
               pass
            elif os.path.isabs(v):
               toadd.append(os.path.dirname(v))
      elif typ=='inc':
         for v in var.split():
            if v[:2]=='-I':
               toadd.append(v[2:])
      elif typ=='exact':
         for v in var.split():
            if os.path.isabs(v):
               toadd.append(v)
      else:
         raise SetupError, _('unexpected type %s') % repr(typ)
      for elt in toadd:
         if not elt in self.paths[typ]:
            self.paths[typ].append(elt)
